/*
 *
 * Difficulty (Update)
 *
 */

const events = require('events');
const utils = require('./utils');
const circularBuffer = require( '@stdlib/utils-circular-buffer' );

////////////////////////////////////////////////////////////////////////////////

// Main Difficulty Function
const Difficulty = function(port, difficulty) {

  const _this = this;
  _this.options = difficulty;

  const sampleSize = 30;
  let lastTs, optimalDiff, cb;

  // Cleanup diff
  this.disconnect = function(client) {
      lastTs = 0;
      cb.clear();
  }

  // Update Difficulty on Share Submission
  this.updateDifficulty = function(client) {
    const ts = (Date.now() / 1000) | 0;

    // We just getting started bois
    if (!lastTs) {
        lastTs = ts;
        cb = circularBuffer(sampleSize);
        optimalDiff = _this.options.initial;
        return;
    }

    // Bail-out early if we're still waiting on difficulty update
    if(optimalDiff != client.difficulty) {
        lastTs = ts;
        return;
    }

    // Update last recorded share timestamp
    cb.push(ts - lastTs);
    lastTs = ts;

    // Keep sampling baby
    if(cb.count < sampleSize) {
        return;
    }

    const coef = _this.options.targetTime / (cb.toArray().reduce((a,b) => a + b, 0) / cb.count);

    // Bail-out if we're good
    if((coef < (1 + _this.options.variance)) && (coef > (1 - _this.options.variance))) {
        return;
    }

    // We're out of "the zone", let's try to fix that
    // That "should"(tm) be the right new difficulty
    optimalDiff = utils.toFixed(client.difficulty * coef, 8);

    // Make sure we don't swing to hard after the initial correction
    if(optimalDiff > (client.difficulty * (1 + _this.options.variance)) && client.difficulty != _this.options.initial) {
        optimalDiff = utils.toFixed(client.difficulty * (1 + _this.options.variance), 8);
    } else if(optimalDiff < (client.difficulty * (1 - _this.options.variance)) && client.difficulty != _this.options.initial) {
        optimalDiff = utils.toFixed(client.difficulty * (1 - _this.options.variance), 8);
    }

    // Ensure we don't go higher than legal maximum and minimum
    if(optimalDiff > _this.options.maximum) {
        optimalDiff = _this.options.maximum;
    } else if(optimalDiff < _this.options.minimum) {
        optimalDiff = _this.options.minimum;
    }

    // Reset buffers then emit new optimal difficulty
    cb.clear();
    _this.emit('newDifficulty', client, optimalDiff);
  };

  // Manage Individual Clients
  this.manageClient = function(client) {
    const stratumPort = client.socket.localPort;
    if (stratumPort != port) {
      console.error('Handling a client which is not of this vardiff?');
    }
    client.on('submit', () => _this.updateDifficulty(client));
    client.on('socketDisconnect', () => _this.disconnect(client));
  };
};

module.exports = Difficulty;
Difficulty.prototype.__proto__ = events.EventEmitter.prototype;
